/*
 *  Digital Audio (PCM) abstraction layer
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#define ULTRA_PCM_ZERO_RESERVED	sizeof( struct ULTRA_PCM_HARDWARE )

#define ULTRA_PCM_PLAYBACK	0
#define ULTRA_PCM_RECORD	1

#define ULTRA_PCM_LFLG_NONE	0x0000
#define ULTRA_PCM_LFLG_PLAY	0x0001
#define ULTRA_PCM_LFLG_RECORD	0x0002
#define ULTRA_PCM_LFLG_BOTH	(ULTRA_PCM_LFLG_PLAY|ULTRA_PCM_LFLG_RECORD) 
#define ULTRA_PCM_LFLG_REALTIME 0x0100
#define ULTRA_PCM_LFLG_VALID	0x4000
#define ULTRA_PCM_LFLG_USED	0x8000

#define ULTRA_PCM_DEFAULT_RATE	8000

#define ULTRA_PCM_MODE_MULTI	0x00000010	/* set - multitrack (1-32) operation enabled */
#define ULTRA_PCM_MODE_VALID	0x00000080	/* unset = not valid, set = valid */
#define ULTRA_PCM_MODE_U	0x00000100	/* unset = signed, set = unsigned */
#define ULTRA_PCM_MODE_16	0x00000200	/* unset = 8 bit , set = 16 bit   */
#define ULTRA_PCM_MODE_BIG	0x00000400	/* big endian (16-bit mode) */
#define ULTRA_PCM_MODE_ULAW	0x00000800	/* mu-Law */
#define ULTRA_PCM_MODE_ALAW	0x00001000	/* a-Law */
#define ULTRA_PCM_MODE_ADPCM	0x00002000	/* IMA ADPCM 4:1 */
#define ULTRA_PCM_MODE_MPEG	0x00004000	/* MPEG 1/2 */
#define ULTRA_PCM_MODE_TYPE	0xffffff00	/* bitmask */

#define ULTRA_PCM_FLG_NONE	0x00000000
#define ULTRA_PCM_FLG_ENABLE	0x00000001	/* enable trigger */
#define ULTRA_PCM_FLG_SLEEP	0x00000002	/* channel is in sleep state - need wakeup */
#define ULTRA_PCM_FLG_ABORT	0x00000004	/* user abort */
#define ULTRA_PCM_FLG_NONBLK	0x00000008	/* non block I/O */
#define ULTRA_PCM_FLG_MMAP	0x00000010	/* mmaped access */
#define ULTRA_PCM_FLG_DMAOK	0x00000020	/* DMA is now allocated */
#define ULTRA_PCM_FLG_NEUTRAL	0x00000040	/* erase DMA buffer (fill with neutral byte) */
#define ULTRA_PCM_FLG_BUFFERS	0x00000080	/* compute new sizes for buffers */
#define ULTRA_PCM_FLG_TRIGGER	0x00000100	/* trigger on/off */
#define ULTRA_PCM_FLG_TRIGGER1	0x00000200	/* prepare ok */
#define ULTRA_PCM_FLG_TRIGGERA	0x00000300	/* both above flags */
#define ULTRA_PCM_FLG_SYNC	0x00000400	/* synchronize playback/record */
#define ULTRA_PCM_FLG_REALTM	0x00000800	/* need realtime... */

#define ULTRA_PCM_HW_AUTODMA	0x00000001	/* hardware supports auto dma - good */

#define ULTRA_PCM_LOCKZERO( channel ) \
  (channel -> block_lock = -1)
#define ULTRA_PCM_LOCK( channel, block ) \
  (channel -> block_lock = block)
#define ULTRA_PCM_ISLOCK( channel, block ) \
  (channel -> block_lock >= 0 && channel -> block_lock == block)

typedef struct ULTRA_PCM_CHANNEL ultra_pcm_channel_t;
typedef struct ULTRA_PCM ultra_pcm_t;

struct ULTRA_PCM_HARDWARE {
  /* -- these values aren't erased -- */
  void *private_data;		/* private structure */
  void (*private_free)( void *private_data );
  unsigned int discarded;	/* discarded blocks... */
  unsigned int xflow;		/* under/overflow */
  /* -- must be filled with low-level driver */
  unsigned short flags;		/* see to ULTRA_PCM_HW_XXXX */
  unsigned int formats;		/* supported formats... */
  unsigned short min_fragment;	/* minimal fragment... */
  unsigned short max_rate;	/* maximal rate... */
  unsigned short max_voices;	/* maximal voices... */
  /* -- low-level functions -- */
  int (*open)( ultra_pcm_t *pcm );
  void (*close)( ultra_pcm_t *pcm );
  void (*prepare)( ultra_pcm_t *pcm, unsigned char *buffer, unsigned int size, unsigned int offset, unsigned int count );
  void (*trigger)( ultra_pcm_t *pcm, int up );
  unsigned int (*pointer)( ultra_pcm_t *pcm, unsigned int used_size );
  void (*dma)( ultra_pcm_t *pcm, unsigned char *buffer, unsigned int offset, unsigned char *user, unsigned int count );
  void (*dma_neutral)( ultra_pcm_t *pcm, unsigned char *buffer, unsigned offset, unsigned int count, unsigned char neutral_byte );
};

struct ULTRA_PCM_CHANNEL {
  /* -- format -- */
  unsigned short voices;	/* or channels 1-32 */
  unsigned int mode;
  unsigned int format;
  unsigned int rate;
  unsigned int requested_block_size;
  unsigned int requested_blocks;
  unsigned int requested_subdivision;
  unsigned int processed_bytes;
  unsigned int interrupts;
  /* -- physical/flags -- */
  unsigned int flags;
  unsigned int used_size;	/* used size of DMA buffer */
  unsigned char neutral_byte;
  struct ULTRA_STRU_DMA *_dma;	/* pointer to dma structure */
  unsigned char *hidden_buffer;	/* pointer to hidden DMA buffer - GUS GF1 PCM */
  unsigned short dma;
  unsigned int size;		/* size of DMA buffer */
  unsigned char *buffer;	/* pointer to DMA buffer */
  /* -- logical blocks -- */
  unsigned short blocks;	/* number of blocks (2-128) */
  unsigned int block_size;	/* size of one block */
  unsigned short used;		/* number of used blocks */
  unsigned int frag_size;	/* size of partly used block */
  unsigned short head;		/* fill it... */
  unsigned short tail;		/* remove it... */
  int block_lock;		/* locked block... */
  /* -- hardware -- */
  struct ULTRA_PCM_HARDWARE hw;
};

struct ULTRA_PCM {
  ultra_card_t *card;		/* ok, must be */
  unsigned int flags;
  unsigned short mask;
  unsigned int info_flags;
  char info_name[ 32 ];
  struct ULTRA_PCM_CHANNEL playback;
  struct ULTRA_PCM_CHANNEL record;
  SLEEP_DEFINE( playback )
  SLEEP_DEFINE( record )
};

extern void ultra_pcm_interrupt_playback( ultra_pcm_t *pcm );
extern void ultra_pcm_interrupt_record( ultra_pcm_t *pcm );

extern void ultra_pcm_playback_dma( ultra_pcm_t *pcm,
			            unsigned char *buffer, unsigned int offset,
			            unsigned char *user, unsigned int count );
extern void ultra_pcm_playback_dma_neutral( ultra_pcm_t *pcm,
				            unsigned char *buffer, unsigned int offset,
				            unsigned int count,
				            unsigned char neutral_byte );
extern void ultra_pcm_record_dma( ultra_pcm_t *pcm,
			          unsigned char *buffer, unsigned int offset,
			          unsigned char *user, unsigned int count );

extern int ultra_pcm_dma_alloc( ultra_card_t *card, ultra_pcm_channel_t *pchn, int dma_number, char *ident, int lock );
extern int ultra_pcm_dma_free( ultra_card_t *card, ultra_pcm_channel_t *pchn, int dma_number, int lock );

/*
 *  Registering
 */

extern ultra_pcm_t *ultra_pcm_new_device( ultra_card_t *card );
extern int ultra_pcm_free( ultra_pcm_t *pcm );
extern int ultra_pcm_register( ultra_pcm_t *pcm, int pcm_device );
extern int ultra_pcm_unregister( ultra_pcm_t *pcm );
